

#include                <Arduino.h>
#include                <Base64.h>
#include                <avr/pgmspace.h>
#include                <SoftwareSerial.h>
#include                <TinyGPS++.h>


// ----------------------------------------------------------------------------
// Constants
// ----------------------------------------------------------------------------

const bool              debugDefault    = false;            // default debugmode
const unsigned long     xmitUdpTimeout  = 60UL;             // secs between xmissions

// ----------------------------------------------------------------------------
// structs
// ----------------------------------------------------------------------------

struct udpRecT {
    double              latitude;
    double              longitude;
    int                 speed;
    int                 course;
    char                unused;                             // base64_decode adds '\0' to end of rec
    };

// ----------------------------------------------------------------------------
// global variables
// ----------------------------------------------------------------------------

bool                    debugMode       = debugDefault;
TinyGPSPlus             gps;
char                    s[80];                              // generic buffer
SoftwareSerial          ssGprs(7, 8);
SoftwareSerial          ssGps(4, 3);

// ----------------------------------------------------------------------------
// Forward declarations
// ----------------------------------------------------------------------------

void die();                                                 // disp msg, hang processor
void doGpsIo();
void doGpsTest();
void doUdpMsg();                                            // xmit data via UDP
void powerUpSms();                                          // make sure SMS card is up
char * readln(                                              // read bytes into <s> until <return> or <maxlen>
    char                *s,
    int                 maxlen,
    int                 timeout         = 1
    );
char * readRec(                                             // read exactly <len> bytes then append '\0'
    char                *s,
    int                 len,
    int                 timeout         = 1
    );
void reboot();
void transmit(                                              // xmit <s> to GPRS
    char *              s
    );
void transmitAT(                                            // send AT cmd, if OK not returned, die
    char *              s,
    int                 timeout         = 1
    );
void transmitAT_P(                                          // send AT cmd, if OK not returned, die
    PGM_P               s,
    int                 timeout         = 1
    );
bool waitfor(                                               // waitfor <t> secs for <s> from GPRS
    char *              s,
    int                 timeout         = 1
    );

// ----------------------------------------------------------------------------
// accept each character from GPS, and send to the GPS decoder

void doGpsIo(
    ) {

char                    c;

while (ssGps.available() > 0) {
    c = ssGps.read();
//  Serial.print(c);                                        // uncomment for debugging
    gps.encode(c);
    }

} // doGpsIo

// ----------------------------------------------------------------------------
// Displays current gps data

void doGpsTest(
    )  {

if (debugMode)
    Serial.println("--------------------");

if (!gps.location.isValid()) {
    Serial.println("no GPS location fix");
    return;
    }

if (gps.charsProcessed() < 10) {
  Serial.println(F("WARNING: No GPS data.  Check wiring."));
  return;
  }

Serial.print(F("https://www.google.com/maps/?q="));
if (gps.location.age() > 5000)
    Serial.print(F("xxx.xxxxxx,xxx.xxxxxx"));
else {
    Serial.print(gps.location.lat(), 6);
    Serial.print(",");
    Serial.print(gps.location.lng(), 6);
    }
Serial.print(F("; "));

if (gps.speed.age() > 5000)
    Serial.print(F("xxx"));
else
    Serial.print(gps.speed.mph(), 0);
Serial.print(F(" MPH;"));

Serial.print(F("Course: "));
if (gps.course.age() > 5000)
    Serial.print(F("xxx"));
else
    Serial.print(gps.course.deg(), 0);
Serial.println(F(";"));

if (debugMode) {
    Serial.print(F("  Sentences that failed checksum="));
    Serial.println(gps.failedChecksum());
    }

if (ssGps.overflow())
    Serial.println(F("ssGps buffer overflowed"));

} // doGpsTest

// ----------------------------------------------------------------------------
// send GPS data via UDP

void doUdpMsg(
    ) {

int                     enc_len;
char                    s[40];
char                    s_enc[40];
char                    s_len[10];
udpRecT                 udpRec;

if (debugMode)
    Serial.println("@doUdpMsg");

// setup encoded string of gps coords

udpRec.latitude     = gps.location.lat();
udpRec.longitude    = gps.location.lng();
udpRec.speed        = gps.speed.mph();
if (gps.speed.mph() < 1.0)
    udpRec.course   = 0;
else
    udpRec.course   = gps.course.deg();

if (debugMode) {
    Serial.print("  ");
    Serial.print("lat: ");
    Serial.print(udpRec.latitude);
    Serial.print("; long: ");
    Serial.print(udpRec.longitude);
    Serial.print("; speed: ");
    Serial.print(udpRec.speed);
    Serial.print("; course: ");
    Serial.print(udpRec.course);
    Serial.println();
    }

enc_len = base64_encode(s_enc, (char*)&udpRec, sizeof(udpRec)-1);

if (debugMode) {
Serial.print("Xmit coordinates in base64: ");
Serial.print(s_enc);
Serial.println();
}

// xmit coords
strncpy(s, "AT+CIPSEND=", sizeof(s));
itoa(enc_len, s_len, 10);
strncat(s, s_len, sizeof(s));

transmit(s);
if (!waitfor(">"))
    die();

transmit(s_enc);
if (!waitfor("OK\r\n"))
    die();

} // doUdpMsg

// ----------------------------------------------------------------------------
// suspend operations

void die(
    ) {

Serial.print(F("program aborted"));
delay(5000UL);
reboot();

} // die

// ----------------------------------------------------------------------------
// turn the GPRS unit on.
// At the moment I don't have it wired such that this will actually work.
// It will just wait until the GPRS is turned on.

void powerUpSms(
    ){

const int               onModulePin     = 9;

if (debugMode)
    Serial.println(F("@powerUpSms"));

// checks if the module is started
ssGprs.println("AT");                                       // don't use transmit-if mdm down, there will be no echo
if (waitfor("OK\r\n"))
    return;

if (debugMode)
    Serial.println(F("  no answer, pulsing power pin."));

// power on pulse
digitalWrite(onModulePin,HIGH);
delay(3000UL);
digitalWrite(onModulePin,LOW);

// waits for an answer from the module
while (true) {
    ssGprs.println("AT");
    if (waitfor("OK\r\n"))
        return;
    delay(1000UL);
    }

if (debugMode)
    Serial.println(F("  modem is on."));

} // powerUpSms

// ----------------------------------------------------------------------------
// read bytes into <s> until <return> or <maxlen>

char * readln(                                              // read <s> from GPRS until <return>
    char                *s,
    int                 maxlen,
    int                 timeout
    ) {

char                    c;                                  // byte read from gprs
int                     i;
unsigned long           timeToPop;

if (debugMode) {
    Serial.print(F("@readln, maxlen="));
    Serial.print(maxlen);
    Serial.println();
    }

timeToPop = millis() + (unsigned long)timeout * 1000UL;

i = 0;                                                      // byte in <s> being tested
while (true) {

    if (i > maxlen - 1) {
        s[i] = '\0';
        return s;
        }

    if (millis() > timeToPop) {
        s[i] = '\0';
        return s;
        }

    if (ssGprs.available() > 0) {
        c = ssGprs.read();
        if (c == '\r') {
            continue;
            }
        else if (c == '\n') {
            s[i] = '\0';
            return s;
            }
        else {
            s[i] = c;
            i    = i + 1;
            }
        }

    } // while

} // readln

// ----------------------------------------------------------------------------
// read fixed # of bytes <len> into s. then append '\0' to s. if timeout
// occurs, <s> will contain whatever was read up to the timeout.

char * readRec(                                             // read exactly <len> bytes then append '\0'
    char                *s,
    int                 len,
    int                 timeout
    ) {

char                    c;                                  // byte read from gprs
int                     i;
unsigned long           timeToPop;

if (debugMode) {
    Serial.print(F("@readRec, len="));
    Serial.print(len);
    Serial.print(F("; timeout="));
    Serial.print(timeout);
    Serial.println();
    }

timeToPop = millis() + (unsigned long)timeout * 1000UL;

i = 0;                                                      // byte in <s> being tested
while (true) {

    if (i == len) {
        s[i] = '\0';
        if (debugMode) {
            Serial.print(F("  read: "));
            Serial.print(s);
            Serial.println();
            }
        return s;
        }

    if (millis() > timeToPop) {
        s[i] = '\0';
        if (debugMode) {
            Serial.print(F("  Timer Popped. read: "));
            Serial.print(s);
            Serial.println();
            }
        return s;
        }

    if (ssGprs.available() > 0) {
        c = ssGprs.read();
        s[i] = c;
        i    = i + 1;
        }

    } // while

} // readRec

//-----------------------------------------------------------------------------
// reboot processor

void reboot(
    ) {

delay(500L);
asm volatile ("  jmp 0");

} //reboot

// ----------------------------------------------------------------------------
// xmit string <s> to GPRS, then sends <CR>

void transmit(
    char *              s
    ) {

char                    c;
unsigned long           timeout;

if (debugMode) {
    Serial.print(F("@transmit, s="));
    Serial.print(s);
    Serial.println();
    }

while (ssGprs.available() > 0)                              // flush buffer
    ssGprs.read();

ssGprs.print(s);
ssGprs.print('\r');

if (debugMode)
    Serial.print(F("  flushed: "));
timeout = millis() + 5000U;
while (true) {                                              // flush buffer of AT command
    if (millis() > timeout) {
        if (debugMode) {
            Serial.println(F("  timeout in transmit function."));
            die();
            }
        return;
        }
    if (ssGprs.available() > 0) {
        c = ssGprs.read();
        if (debugMode) {
            if (c >= ' ')
                Serial.print(c);
            else if (c == '\n')
                Serial.println(F("\\n"));
            else if (c == '\r')
                Serial.print(F("\\r"));
            }
        if (c == '\n')
            break;
        }

    } // while

} // transmit

// ----------------------------------------------------------------------------
// Transmit <s> which is a simple AT cmd that expects only an OK response
// failure to get OK will kill MPU

void transmitAT(
    char *              s,
    int                 timeout
    ) {

if (debugMode) {
    Serial.print(F("@transmitAT: "));
    }

transmit(s);

if (!waitfor("OK\r\n", timeout)) {
    Serial.print(F("transmitAT failure: cmd="));
    Serial.print(s);
    Serial.println();
    die();
    }

} // transmit AT

// ----------------------------------------------------------------------------
// PROGMEM version of transmitAT

void transmitAT_P(                                          // send AT cmd, if OK not returned, die
    PGM_P               s,
    int                 timeout
    ) {

char *                  st;

st = (char*) malloc(strlen_P(s));
strcpy_P(st, s);

transmitAT(st,timeout);

free(st);

} // transmitAT_P

// ----------------------------------------------------------------------------
// waitfor a string to be returned. Return false if string not found

bool waitfor(
    char *              s,                                  // string to wait for
    int                 timeout                             // secs to wait
    ) {

char                    c;                                  // byte read from gprs
int                     i;
unsigned long           timeToPop;

if (debugMode) {
    Serial.print(F("@waitfor, str="));
    for (i = 0; i < strlen(s); i = i + 1) {
        if (s[i] == '\n')
            Serial.print(F("\\n"));
        else if (s[i] == '\r')
            Serial.print(F("\\r"));
        else if (s[i] < ' ')
            Serial.print(F("."));
        else
            Serial.print(s[i]);
        } // for
    Serial.print(F(". timeout="));
    Serial.print(timeout);
    Serial.println();
    }

if (strlen(s) == 0)
    return false;

timeToPop = millis() + 1000UL * (unsigned long)timeout;     // when timer will pop

i = 0;                                                      // byte in <s> being tested
while (true) {

    if (millis() > timeToPop) {
        return false;
        }

    if (ssGprs.available() > 0) {
        c = ssGprs.read();
        if (s[i] == c) {
            i = i + 1;                                      // move on to the next byte to test
            if (i >= strlen(s)) {                           // if no bytes left, we have a match
                return true;
                }
            }
        else {
            i = 0;                                          // match failed, start over
            }
        }

    } // while

} // waitfor

// ----------------------------------------------------------------------------
// setup

void setup(
    ) {

unsigned long           debugTimer;

Serial.begin(9600);                                         // the Serial port of Arduino baud rate.

Serial.print(F("gsmGpsUdpDemo. "));
Serial.print(F(" Compiled: "));
Serial.print(F(__DATE__));
Serial.print(F(" "));
Serial.print(F(__TIME__));
Serial.print(F("."));
Serial.println();
Serial.println();

debugTimer = millis() + 4LU * 1000LU;
Serial.println(F("Press <CR> to enable debugging; otherwise wait 4 secs."));
while (millis() < debugTimer) {
    if (Serial.available()) {
        debugMode = true;
        Serial.println(F("debugging enabled"));
        break;
        }
    } // while

ssGprs.begin(4800);                                         // the GPRS baud rate
ssGps.begin(4800);                                          // GPS baud rate

ssGprs.listen();                                            // listen to GPRS rx pin
powerUpSms();                                               // wait for GPRS to power on

// get GPRS connected to RPI
transmitAT_P(PSTR("ATZ"),5);
transmitAT_P(PSTR("AT+CIPSHUT"), 5);                        // reset GPRS
transmitAT_P(PSTR("AT+CIPMUX=0"), 5);                       // single session
transmitAT_P(PSTR("AT+CIPMODE=0"), 5);                      // non-transparent
transmit("AT+CGATT?");                                      // are we connected?
if (!waitfor("+CGATT: 1\r\n"))
    die();
transmitAT_P(PSTR("AT+CSTT=\"APN\""));                      // setup APN context (user/pass which I don't need)
transmitAT_P(PSTR("AT+CIICR"), 5);                          // bring up GPRS
transmit("AT+CIFSR");                                       // get IP addr
readln(s, sizeof(s));
Serial.print(F("IP: "));
Serial.print(s);
Serial.println();
                                                            // connect to host via udp
transmitAT_P(PSTR("AT+CIPSTART=\"UDP\",\"WWW.YOURDOMAIN.COM\",\"9999\""));
if (!waitfor("CONNECT OK\r\n"))
    die();

randomSeed(analogRead(0));

} // setup

// ----------------------------------------------------------------------------
// loop

void loop(
    ) {

uint8_t                 lastMin             = 0;

ssGps.listen();                                             // listen to GPS rx pin
while (true) {

    doGpsIo();                                              // decode next GPS character

    // millis cannot be trusted due to software serial, so I simply xmit as the minute changes
    if (gps.time.isValid() && gps.time.minute() != lastMin) {
        lastMin = gps.time.minute();
        if (debugMode)
            doGpsTest();
        ssGprs.listen();                                    // start listening to GPRS
        doUdpMsg();                                         // check for incoming SMS and respond w/ coordinates
        ssGps.listen();                                     // go back to listening to GPS
        }

    } // while

// we will never really get here

transmitAT_P(PSTR("AT+CIPCLOSE"), 5);
transmitAT_P(PSTR("AT+CIPSHUT"), 5);

} // loop
